/*
Versione Finale:
Smooth shading corretto, (si genera un po' di facce in pi per generare 
correttamente le normali per vertice e si cancellano alla fine. Aggiunti 
anche i lati della striscia.
Aggiunto moto di precessione al posto della rotazione sull'asse x.

*/
#include<stdio.h>
#include<stdlib.h>
#include<GL/glut.h>
#include<math.h>
#include<vector>
#include <vcg/Point3.h>
using namespace vcg;
using namespace std;
// globale per memorizzare l'aspect ratio della finestra correntte; 
// usata da myRedraw, e myReshape
float aspect=1;


class myVertex
{
public :
	myVertex(Point3d &pp,Point3d &nn) {p=pp; n=nn;}
	Point3d p;
	Point3d n;
};

class myFace
{
public:
	myVertex *v[3];
	Point3d n;
};

class myMesh {
public:
	vector< myVertex > vert;
	vector< myFace > face;
	void ComputeFaceNormal();
	void ComputeVertNormal();
};

void myMesh::ComputeFaceNormal()
{
	vector< myFace >::iterator fi;
	for(fi=face.begin();fi!=face.end();++fi)
		(*fi).n= Normal((*fi).v[0]->p,(*fi).v[1]->p,(*fi).v[2]->p);
}

// Da chiamare solo DOPO la ComputeFaceNormal
void myMesh::ComputeVertNormal()
{
	vector< myFace >::iterator fi;
	for(fi=face.begin();fi!=face.end();++fi)
		for(int i=0;i<3;++i)
			(*fi).v[i]->n += (*fi).n;
}


myMesh Ring;

// Inizializzazione dello stato di OpenGL
void myInitGL()
{
	// Abilita il la possibilita' di cambiare il alcune componenti colore del 
	// materiale tramite un semplice glcolor, anzich fare glMaterial(...)
	glColorMaterial(GL_FRONT_AND_BACK,GL_AMBIENT_AND_DIFFUSE);
	glEnable(GL_COLOR_MATERIAL);

	// Abilita il calcolo dell'illuminazione
	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);

	// anche le normali attraversano la pipeline, quindi devono essere  
	// rinormalizzate, prima del calcolo dell'illuminazione, altrimenti 
	// il calcolo dell'illuminazione viene sbagliato.
	glEnable(GL_NORMALIZE); 
	
	// Abilita il test sullo zbuffer e quindi l'eliminazione delle 
	// superfici nascoste
	glEnable(GL_DEPTH_TEST); 	
}


void GenerateStrip()
{
	const float Radius=3;
	const float BlockSide=3;
	int StepNum=36;
	float StepAngleDeg=360.0/StepNum;
	float StepAngleRad=2.0*3.14159265/StepNum;

	double MM[16],MP[16];
	

	//int OldViewPort[4];
	//glGetIntegerv(GL_VIEWPORT,OldViewPort);
	glMatrixMode(GL_PROJECTION);	
	glLoadIdentity();
	glOrtho(-1,1,-1,1,-1,1);
	glGetDoublev(GL_PROJECTION_MATRIX,MP);

	// Inizializzazione della Modelview Matrix
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();	
  	
	// -- Disegno dell'Anello -- 
	// l'anello viene create simmetrico rispetto all'asse z.
	// Per renderlo di moebius basta fare aumentare la rotazione lungo
	// l'asse y, fatta prima di ruotare intorno all'asse z; Tale rotazione 
  // deve essere la meta' di quella intorno all'asse z
	// (cosi' dopo 360 gradi ha fatto 180 gradi)
	
	// Per far si'che il calcolo delle normali interpolate sia corretto si 
	// aggiunge due strisce di triangoli in piu' che vengono tolte alla fine. 
  // alla fine si togliera quindi 4 triangoli all'inizio e quattro alla fine
	Point3d nn(0,0,0);
	for(int i=0;i<StepNum+3;++i)
	{
		glPushMatrix();
		glRotatef(i*StepAngleDeg    , 0,0,1);   // Rotazione Intorno all'asse principale
		glTranslatef(Radius,0,0);
		glRotatef(i*StepAngleDeg/2.0, 0,1,0);   // Avvolgimento della striscia.
	
		// -- Disegno del blocchetto dell'anello -- 
		Point3d pp;
		glGetDoublev(GL_MODELVIEW_MATRIX,MM);
		int vp[4]={-1,-1,2,2};
		
		gluProject(-.8,0,.05,MM,MP,vp,&pp[0],&pp[1],&pp[2]);
		pp[2]=-2.0f*(pp[2]-.5);
		Ring.vert.push_back(myVertex(pp,nn));
		
		gluProject(.8,0,.05,MM,MP,vp,&pp[0],&pp[1],&pp[2]);
		pp[2]=-2.0f*(pp[2]-.5);
		Ring.vert.push_back(myVertex(pp,nn));
		
		gluProject(.8,0,-.05,MM,MP,vp,&pp[0],&pp[1],&pp[2]);
		pp[2]=-2.0f*(pp[2]-.5);
		Ring.vert.push_back(myVertex(pp,nn));
		
		gluProject(-.8,0,-.05,MM,MP,vp,&pp[0],&pp[1],&pp[2]);
		pp[2]=-2.0f*(pp[2]-.5);
		Ring.vert.push_back(myVertex(pp,nn));
		
		glPopMatrix();
	}
  //  a--c
	//  |\ |
	//  | \|
	//  b--d  
	//


	for(int i=0;i<StepNum+2;++i)
	{
		for(int k=0;k<4;++k)
		{
			myFace tmp;

			tmp.v[0]=&(Ring.vert[(i+0)*4+(k+0)%4]); // a 
			tmp.v[1]=&(Ring.vert[(i+0)*4+(k+1)%4]); // b 
			tmp.v[2]=&(Ring.vert[(i+1)*4+(k+1)%4]); // d 
			Ring.face.push_back(tmp);

			tmp.v[0]=&(Ring.vert[(i+0)*4+(k+0)%4]); // a
			tmp.v[1]=&(Ring.vert[(i+1)*4+(k+1)%4]); // d
			tmp.v[2]=&(Ring.vert[(i+1)*4+(k+0)%4]); // c
			Ring.face.push_back(tmp);
		}
	}

	Ring.ComputeFaceNormal();
	Ring.ComputeVertNormal();

  // cancello la prima e l'ultima striscia di facce che avevo 
	// messo solo per calcolare in maniera corretta le normali per vertice.
	// Nota ora ci sono in Ring.vert alcuni vertici che non sono riferiti 
	// da alcuna faccia
	Ring.face.erase(Ring.face.begin(),Ring.face.begin()+8);
	Ring.face.erase(Ring.face.end()-8,Ring.face.end());
}

// Funzione principale di aggiornamento dello schermo
void myRedrawFunc()
{
	int i;
	
	// Gestione dell'animazione
	float AngularVelocity = 120; // degree per sec;
	float ElapsedSecs = float(glutGet(GLUT_ELAPSED_TIME))/1000.0;
	float CurAngleDeg = AngularVelocity * ElapsedSecs;
	float CurAngleRad = CurAngleDeg/180.0*3.14159265;
	
	//Parametri dell'Anello
	const float Radius=3;
	const float BlockSide=3;
	int StepNum=24;
	float StepAngleDeg=360.0/StepNum;
	float StepAngleRad=2.0*3.14159265/StepNum;
	float InternalRotationSpeed=0.33;
	float BallSpeed=0.5;

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glMatrixMode(GL_PROJECTION);	
	glLoadIdentity();
	gluPerspective(45,aspect,5,15);

	// Inizializzazione della Modelview Matrix
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();	
  
	// World To Camera Transformation
	gluLookAt(0,0,10,0,0,0,0,1,0);
	glRotatef( -90,0,1,0);  
	glRotatef( 90,1,0,0);  	
 	
	// moto di precessione: una rotazione il cui asse ruota intorno all'asse z
	glRotatef( 10,cos(ToRad(-45+CurAngleDeg*.5)),sin(ToRad(-45+CurAngleDeg*.5)),0);  

	if(Ring.vert.empty())
		GenerateStrip();
	//glDisable(GL_LIGHTING);
	glColor3f(1,.2,.2);
	
	glBegin(GL_TRIANGLES);
	
	for(int i=0;i<Ring.face.size();++i)
	{
		glNormal(Ring.face[i].v[0]->n);
		glVertex(Ring.face[i].v[0]->p);
		glNormal(Ring.face[i].v[1]->n);
		glVertex(Ring.face[i].v[1]->p);
		glNormal(Ring.face[i].v[2]->n);
		glVertex(Ring.face[i].v[2]->p);
	}
	glEnd();


	// -- Disegno della Pallina che rotola -- 
	// Assomiglia molto al disegno di un singolo blocchetto dell'anello.
	// a cui va aggiunta una rotazione attorno all'asse x per fare l'effetto
	// di rotolamento della pallina sulla striscia
  const float BallRadius=.75;
	glRotatef(CurAngleDeg*BallSpeed,0,0,1);
	glTranslatef(Radius,0,0);
	glRotatef(CurAngleDeg*BallSpeed/2,0,1,0);
		
	glTranslatef(0,0,BallRadius+.05);  // Per spostarla dal piano dell'anello
	glRotatef(-CurAngleDeg*BallSpeed*3,1,0,0); // Rotolamento (la quantita' e' occhio...)
	
	// -- Disegno della pallina --
	// due sfere rosse/blu leggermente scalate
	glColor3f(0,0,1);
	glutSolidSphere(BallRadius,20,10);
	glColor3f(1,0,0);
	glScalef(1.05f,1.05f,.6f);
	glutSolidSphere(BallRadius,20,10);

	glFlush();
  glutSwapBuffers();
}

void myIdle()
{
	glutPostRedisplay();
}

void myReshapeFunc(int wx, int wy)
{
	glViewport(0,0,wx,wy);
	aspect=float(wx)/float(wy);
}
int main(int argc, char *argv[])
{
	glutInitDisplayMode(GLUT_RGBA | GLUT_DEPTH | GLUT_DOUBLE);
	glutInit(&argc, argv);
	glutIdleFunc(myIdle);
	glutCreateWindow("Moebius");
	glutDisplayFunc(myRedrawFunc);
	glutReshapeFunc(myReshapeFunc);
	myInitGL();
	glutMainLoop(); 	//Passare al SO il controllo.
}